# -*- coding: utf-8 -*-
"""
Created on Mon Oct 22 16:44:47 2018

Convenience functions for visualizing radiographic data and
processed data from XRIPL.

@author: Pawel M. Kozlowski
"""

# python modules
import numpy as np
import matplotlib.pyplot as plt


def spatialConversion(img, px2Um, pxRef, umRef, method='vertical'):
    r"""
    Generates spatial extents of image in micrometers, by using
    the image shape, pixel to micrometer conversion factor, and
    a reference position defined in both pixel and micrometer
    coordinates.
    
    img : numpy.ndarray
        Image for getting spatial conversion coordinates.
        
    px2Um : float
        Conversion factor micrometers / pixel for the image.
        
    pxRef : tuple
        Reference position in pixels.
        
    umRef : tuple
        Reference position in micrometers.
        
    method : str
        Marble images are horizontal by default, but we make them
        vertical to match Ranjan SBI paper with shock propagating
        downward. Default shock direction is vertical.
    
    """
    shape = np.shape(img)
    if method == 'vertical':
        # getting spatially calibrated axes. This works for cropped
        # image with shock propagating from top to bottom of image.
        hoMin = (0 - pxRef[0]) * px2Um + umRef[1]
        hoMax = (shape[1] - pxRef[0]) * px2Um + umRef[1]
        vertMin = (0 - pxRef[1]) * px2Um + umRef[0]
        vertMax = (shape[0] - pxRef[1]) * px2Um + umRef[0]
        extent = [hoMin, hoMax, vertMax, vertMin]
    elif method == 'horizontal':
        # getting spatially calibrated axes. This works for full,
        # uncropped image in oriented with shock propagating in
        # horizontal direction from right to left.
        hoMin = -1 * ((0 - pxRef[0]) * px2Um - umRef[0])
        hoMax = -1 * ((shape[1] - pxRef[0]) * px2Um - umRef[0])
        vertMin = (0 - pxRef[1]) * px2Um + umRef[1]
        vertMax = (shape[0] - pxRef[1]) * px2Um + umRef[1]
        extent = [hoMin, hoMax, vertMax, vertMin]
    else:
        raise Exception(f"No such method {method}!")
    return extent


def overlayTube(img, extent, method='vertical'):
    r"""
    Overlays known coordinates of tube based on shock tube dimensions
    onto the given image.
    
    img : numpy.ndarray
        Spatially calibrated radiograph.
        
    extent : list
        Extent coordinates for plotting the spatially calibrated image
        using plt.imshow(). See spatialCovnersion().
        
    method : str
        Orientation of image. This should be the same as used
        in spatialConversion() to get extent.
    """
    plt.figure(figsize=(10,12))
    plt.imshow(img, extent=extent)
    if method == 'vertical':
        # inner tube diameter
        plt.axvline(-250, color='red', ls='--')
        plt.axvline(250, color='red', ls='--')
        # outer tube diameter
        plt.axvline(-300, color='red', ls='--')
        plt.axvline(300, color='red', ls='--')
        plt.xlabel(r'Radial ($\rm \mu m$)')
        plt.ylabel(r'Axial ($\rm \mu m$)')
    elif method == 'horizontal':
        # inner tube diameter
        plt.axhline(-250, color='red', ls='--')
        plt.axhline(250, color='red', ls='--')
        # outer tube diameter
        plt.axhline(-300, color='red', ls='--')
        plt.axhline(300, color='red', ls='--')
        plt.xlabel(r'Axial ($\rm \mu m$)')
        plt.ylabel(r'Radial ($\rm \mu m$)')
    else:
        raise Exception(f"No such method {method}!")
    plt.title('Tube coordinates check')
#    plt.axis([1500, 750, -350, 350])
    plt.show()
    return


def cropImgCalibrated(img,
                      px2Um,
                      pxRef,
                      umRef,
                      xMin,
                      xMax,
                      yMin,
                      yMax,
                      method='horizontal',
                      plots=False):
    r"""
    Crops the image and provides updated extents for spatial
    calibration of the image.
    
    img : numpy.ndarray
        Image to be cropped.
        
    px2Um : float
        Conversion factor micrometers / pixel for the image.
        
    pxRef : tuple
        Reference position in pixels for the uncropped image. This
        is based on a known feature, such as a fiducial or filter edge.
        
    umRef : tuple
        Reference position in micrometers. This is the same reference
        used for pxRef, but with the known target coordinates from
        metrology/design.
        
    xMin : int
        Lower bound pixel for cropping in x-direction.
        
    xMax : int
        Upper bound pixel for cropping in x-direction.
        
    yMin : int
        Lower bound pixel for cropping in y-direction.
        
    yMax : int
        Upper bound pixel for cropping in y-direction.
        
    method : str
        Direction of shock propagation in the image. Default shock
        direction is horizontal.
        
    plots : bool
        Flag for plotting cropped image with spatially calibrated axes.
        Default is False.
    """
    # cropping the image
    img_crop = img[yMin:yMax, xMin:xMax]
    
    # transforming coordinates for cropped image
    pxRefCrop = (pxRef[0] - xMin, pxRef[1] - yMin)
    
    # spatial calibration for cropped image
    extent_crop = spatialConversion(img=img_crop,
                                    px2Um=px2Um,
                                    pxRef=pxRefCrop,
                                    umRef=umRef,
                                    method=method)
    
    if plots:
        # plotting the cropped image
        plt.imshow(img_crop, extent=extent_crop)
        #plt.scatter(pxRefCrop[0],
        #            pxRefCrop[1],
        #            marker="+",
        #            c="C6",
        #            s=1000)
        plt.scatter(umRef[0],
                    umRef[1],
                    marker="+",
                    c="C6",
                    s=1000)
        plt.xlabel(r'Axial ($\rm \mu m$)')
        plt.ylabel(r'Radial ($\rm \mu m$)')
        plt.title("Cropped")
        plt.show()
    # returns cropped image, extents for applying spatial calibration
    # when plotting the cropped image, and the updated pixel position
    # of the reference point used for calibration.
    return img_crop, extent_crop, pxRefCrop


def rotateImgCalibrated(img,
                        px2Um,
                        pxRef,
                        umRef,
                        method='vertical',
                        plots=False):
    r"""
    Rotates the image 90 degrees in counter clockwise direction, and
    provides updated extents for spatial calibration of the image.
    
    img : numpy.ndarray
        Image to be rotated.
        
    px2Um : float
        Conversion factor micrometers / pixel for the image.
        
    pxRef : tuple
        Reference position in pixels for the unrotated image. This
        is based on a known feature, such as a fiducial or filter edge.
        
    umRef : tuple
        Reference position in micrometers. This is the same reference
        used for pxRef, but with the known target coordinates from
        metrology/design.
        
    method : str
        Direction of shock propagation in the newly rotated image. Default
        shock direction is vertical.
        
    plots : bool
        Flag for plotting rotated image with spatially calibrated axes.
        Default is False.
    """
    # rotating 90 degrees CCW using a transpose and a flip
    img_rot = np.flipud(img.transpose())
    cropShape = np.shape(img_rot)
    # getting updated px position of reference point for the rotated
    # image.
    pxRefRot = (pxRef[1], cropShape[0] - pxRef[0])
    # getting updated spatially calibrated axes extents for the
    # newly rotated image.
    extent_rot = spatialConversion(img=img_rot,
                                   px2Um=px2Um,
                                   pxRef=pxRefRot,
                                   umRef=umRef,
                                   method=method)
    
    if plots:
        # plotting the rotated image
        plt.imshow(img_rot, extent=extent_rot)
        #plt.scatter(pxRefFlip[0],
        #            pxRefFlip[1],
        #            marker="+",
        #            c="C6",
        #            s=1000)
        plt.scatter(umRef[1],
                    umRef[0],
                    marker="+",
                    c="C6",
                    s=1000)
        plt.xlabel(r'Radial ($\rm \mu m$)')
        plt.ylabel(r'Axial ($\rm \mu m$)')
        plt.title("Rotated")
        plt.show()
    return img_rot, extent_rot, pxRefRot


def lineoutsComparison1(imgRaw,
                        imgDenoised,
                        imgCleaned,
                        px,
                        extent=None,
                        showFlag=True):
    r"""
    Plots normalized lineouts for raw, denoised, and cleaned iamges for
    comparison. Plot shows how well noise is removed and how
    well features are preserved.
    
    imgRaw : numpy.ndarray
        Raw radiographic image.
        
    imgDenoised : numpy.ndarray
        Denoised (median filtered) radiographic image.
        
    imgCleaned : numpy.ndarray
        Cleaned (morphologically filtered) radiographic image.
        
    px : int
        Pixel position at which to take the lineouts.
        
    extent : list
        Extents defining spatial axes. Used to form radial axis in um.
        Default is None, which falls back on radial axis in pixels.
        
    showFlag : bool
        Flag for showing the plotted image. Set to False if you want
        to further modify the plot. Default is True.
    """
    # lineout and normalized lineout from raw image
    lineoutY = imgRaw[px, :]
    lineoutY = lineoutY / np.max(lineoutY)
    # lineout and normalized lineout from denoised (median filtered) iamge.
    lineoutYDenoised = imgDenoised[px, :]
    lineoutYDenoised = lineoutYDenoised / np.max(lineoutYDenoised)
    # lineout and normalized lineout from cleaned (morphologically filtered)
    # image.
    lineoutYCleaned = imgCleaned[px, :]
    lineoutYCleaned = lineoutYCleaned / np.max(lineoutYCleaned)
    # forming x-axis basis of lineouts
    if extent:
        lineoutX = np.linspace(start=extent[0],
                               stop=extent[1],
                               num=np.shape(imgRaw)[1])
    else:
       lineoutX = np.arange(np.shape(imgRaw)[1])
      
    # plotting
    plt.plot(lineoutX, lineoutY, label='orig')
    plt.plot(lineoutX, lineoutYDenoised, label='denoised')
    plt.plot(lineoutX, lineoutYCleaned, label='cleaned')
    if extent:
        plt.xlabel(r'Radial ($\rm \mu m$)')
    else:
        plt.xlabel('Radial (px)')
    plt.legend(frameon=False,
               labelspacing=0.001,
               borderaxespad=0.1)
    
    if showFlag:
        plt.show()
    return    


def lineoutsComparison2(imgDenoised,
                        imgFlat,
                        px,
                        extent=None,
                        showFlag=True):
    r"""
    Compares pseudo-flatfielded lineout against denoised lineout.
    Pseudo-flatfielded lineout typically includes cleaning via
    morphological filtering, which can distort the image, whereas
    denoising is done through median filtering, which typically
    preserves features. This means this comparison is important
    for showing whether features are preserved through the morphological
    filtering step and whether the pseudo-flatifield suitably 
    flattens the intensity curve compared to what we expect.
    In addition, this overlays expected tube positions for Marble
    VC 16A target.
    
    imgDenoised : numpy.ndarray
        Denoised (median filtered) radiographic image.
        
    imgFlat : numpy.ndarray
        Pseudo-flatfielded image (using over-blurring).
        
    px : int
        Pixel position at which to take the lineouts.
        
    extent : list
        Extents defining spatial axes. Used to form radial axis in um.
        Default is None, which falls back on radial axis in pixels.
        
    showFlag : bool
        Flag for showing the plotted image. Set to False if you want
        to further modify the plot. Default is True.
    
    """
    # lineout and normalized lineout from denoised (median filtered) iamge.
    lineoutYRot = imgDenoised[px, :]
    lineoutYRot = lineoutYRot / np.max(lineoutYRot)
    # lineout and normalized lineout from pseudo-flatfielded image.
    lineoutYFlat = imgFlat[px, :]
    lineoutYFlat = lineoutYFlat / np.max(lineoutYFlat)
    # forming x-axis basis of lineouts
    if extent:
        lineoutXFlat = np.linspace(start=extent[0],
                                   stop=extent[1],
                                   num=np.shape(imgFlat)[1])
    else:
       lineoutXFlat = np.arange(np.shape(imgFlat)[1])     
    
    # plotting
    plt.plot(lineoutXFlat, lineoutYFlat, label='flat')
    plt.plot(lineoutXFlat, lineoutYRot, label='denoised')
    plt.axvline(-250, color='red', ls='--')
    plt.axvline(250, color='red', ls='--')
    plt.axvline(-300, color='red', ls='--')
    plt.axvline(300, color='red', ls='--')
    if extent:
        plt.xlabel(r'Radial ($\rm \mu m$)')
    else:
        plt.xlabel('Radial (px)')
    plt.legend(frameon=False,
               labelspacing=0.001,
               borderaxespad=0.1)
    
    if showFlag:
        plt.show()
    return lineoutXFlat, lineoutYFlat


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Jan 10 10:10:36 2019

Image segmentation tools for radiographic images.

@author: Pawel M. Kozlowski
"""

# python modules
import numpy as np
from scipy import ndimage as ndi
from skimage import measure
from skimage.morphology import watershed, disk
from skimage.filters import median, rank
from skimage.filters.rank import maximum, minimum
import matplotlib.pyplot as plt

# custom modules
from xripl.contour import nContours
import xripl.pltDefaults


def merge_labels(labels_image,
                 labels_to_merge,
                 label_after_merge):
    r"""
    Utility function for merging different labeled regions.
    """
    labels_map = np.arange(np.max(labels_image) + 1)
    labels_map[labels_to_merge] = label_after_merge
    return labels_map[labels_image]


def nSegments(labels, n):
    """
    Get the n largest area segments from a segmented (labeled) image.
    
    labels: numpy.ndarray
        A segmented 2D image.
        
    n: int
        Number of segments to select.
    """
    # getting region properties of segments
    props = measure.regionprops(labels)
    # getting areas of segments
    areas = np.array([p['filled_area'] for p in props])
    # getting label number corresponding to each segments
    numbers = np.array([p['label'] for p in props])
    # sorting indices of areas list based on area size of segment
    areasIdxSorted = np.argsort(areas)[::-1]
    # Reducing sorted index list to n largest areas
    largestAreasIdxs = areasIdxSorted[:n]
    # filtering segments list using selected indices of n largest segments
    largestSegments = numbers[largestAreasIdxs]
    return largestSegments


def segmentContour(labels, segmentNumber, plots=False):
    """
    Given a segmented image and a selected segment number, obtains the
    longest contour of that segment.
    
    labels: numpy.ndarray
        A segmented 2D image.
        
    segmentNumber: int
        Which labeled segment to choose for obtaining contour.
    
    plots: bool
        Flag for plotting longest contour over binary image of segment.
    """
    # transforming image into binary based on selected label
    binImg = labels == segmentNumber    
    # Getting contours of segment
    contoursSegment = measure.find_contours(binImg, 0)    
    # selecting the longest contour
    contourLongest = nContours(contoursSegment, 1)[0]
    if plots:
        # and plotting longest contour over binary image
        plt.imshow(binImg)
        plt.plot(contourLongest[:, 1], contourLongest[:, 0], linewidth=2)
        plt.show()
    return contourLongest


#%% collection of methods for detection of shock position in radiographs
    
def detectShock(image,
                originalImage=None,
                medianDisk=40,
                gradientDiskMarkers=10,
                gradientThreshold=10,
                gradientDiskWatershed=4,
                morphDisk=10,
                compactness=0,
                plots=True,
                nSegs=3):
    """
    This is a convenience function for applying watershed segmentation using
    regions of low gradient as markers, and using another gradient image
    for processing via watershed.
    
    Steps:
        1st the image is denoised using a median filter
        2nd the gradient of the image is obtained and thresholded to generate
            markers to be used for watershed segmentation.
        3rd Morphological closing operation is applied to markers to make
            the region boundaries thicker and the marker regions more
            continous.
        4th Another gradient image is generated, which will be processed
            via watershed.
        5th The gradient image in step 4 and the markers from step 3 are
            processed using watershed segmentation.
    
    image: numpy.ndarray
        Image to be processed for shock detection.
        
    originalImage: numpy.ndarray
        Original image, in case user want to plot contours and segments
        over a different image. Default is None.
        
    medianDisk: int
        Size of disk used in median filter application.
        
    gradientDiskMarkers: int
        Size of disk used in gradient function for determining watershed
        markers. This makes the gradient image smoother.
        
    gradientThreshold: int
        Maximum size of gradient to be used for selecting watershed
        markers.
        
    gradientDiskWatershed: int
        Size of disk used in gradient function for generating image to be
        processed by watershed. This makes the gradient image smoother.
        
    morphDisk: int
        Size of disk used in morphological closing of markers image. This
        cleans up the markers into more continuous regions. If zero, then
        this step is skipped.
        
    compactness: float
        Compactness parameter for watershed. 
        See skimage.morphology.watershed().
        
    plots: bool
        Flag for plotting results
        
    nSegs: int
        Obtains contours of the largest segments by filled area.
    """
    if not type(originalImage) == np.ndarray:
        originalImage = image
    # normalizing image and original
#    image_norm = image / np.max(img)
    original_norm = originalImage / np.max(originalImage)
    denoised = median(image, disk(medianDisk))
    # normalizing
    denoised_norm = denoised / np.max(denoised)
    # find continuous region (low gradient - where less than gradientThreshold
    # for this image) --> markers
    # disk(gradientDiskMarkers) is used here to get a more smooth image
    rankGrad = rank.gradient(denoised_norm, disk(gradientDiskMarkers))
    markers = rankGrad < gradientThreshold
    
    # morphological operations on markers to close some gaps that blend
    # the shock with the background
    if morphDisk != 0:
        closing = maximum(minimum(markers, disk(morphDisk)), disk(morphDisk))
    else:
        closing = markers
    
    # plotting morphologically closed markers
    plt.imshow(closing)
    plt.title('Markers - morphologically closed')
    plt.show()
    
    # labeling markers
    markers = ndi.label(closing)[0]
    
    # local gradient (disk() is used to keep edges thin). This gradient is
    # used directly in the watershed.
    gradient2 = rank.gradient(denoised_norm, disk(gradientDiskWatershed))
    
    # process using watershed
    labels = watershed(gradient2, markers, compactness=compactness)
    
    # getting contours for largest area segments
    contoursList = []
    largestSegments = nSegments(labels=labels, n=nSegs)
    for segmentNumber in largestSegments:
        # get longest contour for each segment
        contourLongest = segmentContour(labels=labels,
                                        segmentNumber=segmentNumber,
                                        plots=False)
        # save the contour to list
        contoursList.append(contourLongest)
    
    # plotting results
    if plots:
        # denoised image
        fig, ax = plt.subplots(figsize=(12, 8))
        plt.imshow(denoised_norm)
        plt.colorbar()
        plt.title('Denoised')
        plt.show
        
        # comparisons between original, gradient, markers, and segmented
        # images.
        fig, axes = plt.subplots(nrows=2,
                                 ncols=2,
                                 figsize=(12, 12),
                                 sharex=True,
                                 sharey=True)
        ax = axes.ravel()
        
        ax[0].imshow(original_norm)
        ax[0].set_title("Original")
        
        ax[1].imshow(gradient2, cmap=plt.cm.tab20)
        ax[1].set_title("Local Gradient")
        
        ax[2].imshow(markers, cmap=plt.cm.tab20)
        ax[2].set_title("Markers")
        
        ax[3].imshow(original_norm, cmap=plt.cm.gray)
        ax[3].imshow(labels, cmap=plt.cm.tab20, alpha=.3)
        ax[3].set_title("Segmented")
        fig.tight_layout()
        plt.show()
        
        # Bigger image of segmentation
        fig, ax = plt.subplots(figsize=(12, 8))
        plt.imshow(original_norm, cmap=plt.cm.gray)
        plt.imshow(labels, cmap=plt.cm.tab20, alpha=.5)
        plt.title('Segments')
        plt.tight_layout()
        plt.show()
        
        # plotting contours
        fig, ax = plt.subplots(figsize=(12, 8))
        plt.imshow(original_norm, cmap=plt.cm.gray, vmax=0.3)
        plt.colorbar()
        for contour in contoursList:
            plt.plot(contour[:, 1], contour[:, 0], linewidth=2)
        plt.title('Contours of segments')
        plt.show()
            
        
#        # getting contours
#        contoursList = []
#        vals = np.arange(np.max(labels)) + 1
#        fig, ax = plt.subplots(figsize=(12, 8))
#        plt.imshow(original_norm, cmap=plt.cm.gray)
#        plt.colorbar()
#        for n in vals:
#            contours = measure.find_contours(labels, n)
#            if contours:
#                contour = contours[0]
#                contoursList.append(contour)
#                plt.plot(contour[:, 1], contour[:, 0], linewidth=2)
#        plt.title('Contours of segments')
#        plt.show()
    
    return labels, contoursList, gradient2, markers


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Jun  3 14:51:45 2020

@author: Pawel M. Kozlowski
"""

import numpy as np
import matplotlib.pyplot as plt
import scipy.signal as sig

import xripl.pltDefaults
from xripl.reader import shotRadiograph
from xripl.clean import enhanceRadiograph
from xripl.visualizations import cropImgCalibrated, rotateImgCalibrated


def lineoutMinima(lineoutX,
                  lineoutY,
                  window_length=51,
                  polyOrder=3,
                  peakWidth=50,
                  plotsFlag=False):
    r"""
    Given lineout should be lineout from flattened radiograph
    image. This function smoothes the lineout using a Savitzky-Golay
    filter, and finds the 2 deepest minima in the lineout. These minima
    should correspond to the inner diameter edge of the target tube.
    
    lineoutX : numpy.ndarray
        1D array of x-axis points of the lineout from which we
        want to get minima.
        
    lineoutY : numpy.ndarray
        1D array of y-axis points of the lineout from which we want to
        get minima. This should be a radial lineout through the tube target.
        Finds minima in intensity, which should correpsond
        to the inner diameter edge of the tube.
    
    window_length : int
        Window length for Savitzky-Golay filter. Must be odd numbered.
        Default is window that is 51 points long.
    
    polyOrder : int
        Polynomial order for Savitzky-Golay filter. Default is 3 (cubic).
        
    peakWidth : int
        Minimum number of points required in peak width to select
        the peak as a minima. Default is 50 pts wide.
        
    plotsFlag : bool
        Flag for plotting smoothed lineout over input lineout. Also
        plots minima points over smoothed lineout. Default is False.
    """
    # Savitzky-Golay smoothing the signal to prevent noise from
    # distorting peak finder from finding minima corresponding to
    # inner edge of target tube.
    lineoutSmoothed = sig.savgol_filter(lineoutY,
                                        window_length,
                                        polyOrder)
    
    # Inverting the lineout so that peak finder gets minima, since
    # peak finder is made to look for maxima.
    inverted = -lineoutSmoothed
    # Finding peaks with minimum width of peakWidth
    peaks, properties = sig.find_peaks(inverted,
                                       width=peakWidth)
    
    if len(peaks) < 2:
        print("Found fewer than 2 peaks. Trying with smaller width.")
        widthMod = int(peakWidth * 1/2)
        print(f"Trying width of {widthMod} in lieu of {peakWidth}.")
        peaks, properties = sig.find_peaks(inverted,
                                           width=widthMod)
        # setting plot to true for inspection
        plotsFlag = True
    
    # fetching x and y coordinates of peaks
    minimaPeaksX = lineoutX[peaks]
    minimaPeaksY = lineoutSmoothed[peaks]
    
    # filtering list of found peaks to the 2 smallest ones in terms of
    # intensity.
    sortIdxs = np.argsort(minimaPeaksY)
    selectIdxs = sortIdxs[:2]
    minimaPeaksXFilt = minimaPeaksX[selectIdxs]
    minimaPeaksYFilt = minimaPeaksY[selectIdxs]
    
    # getting pixel positions of minima
    minimaIdxs = peaks[selectIdxs]
    
    if plotsFlag:
        # plotting the input lineout against the smoothed lineout to
        # ensure that features were preserved and not over-smoothed
        plt.plot(lineoutX, lineoutY, label='flat')
        plt.plot(lineoutX, lineoutSmoothed, label='smoothed')
        plt.xlabel(r'Radial ($\rm \mu m$)')
        plt.legend(frameon=False,
                   labelspacing=0.001,
                   borderaxespad=0.1)
        plt.show()
        
        # plotting 2 smallest minima over smoothed lineout.
        plt.plot(lineoutX, lineoutSmoothed)
        plt.scatter(minimaPeaksXFilt,
                    minimaPeaksYFilt,
                    c='red',
                    marker='o',
                    s=200)
        plt.xlabel(r'Radial ($\rm \mu m$)')
        plt.show()
    return minimaIdxs, minimaPeaksXFilt, minimaPeaksYFilt


def tubeCenterPx(minimaIdxs, lineoutY=None, plotsFlag=False):
    r"""
    Given the pixel positions of the tube edges, obtains the tube center
    in pixels for a tube target.
    
    minimaIdxs : tuple
        A tuple of ints corresponding to the pixel positions of the
        two minima in intensity found in the radiograph by lineoutMinima().
        These minima correspond to the inner edge of the target tube.
    
    lineoutY : numpy.ndarray
        Intensity values of the lineout from which tube edges were found.
        This is only used for plotting. Default is None.
    
    plotsFlag : bool
        Plots the lineout and overlays the pixel position of the tube edges
        and center. Default is False.
    """
    # checking that two minima were passed
    if np.shape(minimaIdxs) == (2,):
        # finding the lower and higher pixel values
        idxLower = np.min(minimaIdxs) 
        idxUpper = np.max(minimaIdxs)
        center = (idxUpper + idxLower) / 2
    else:
        raise Exception (f"Expected 2 minima, but got {len(minimaIdxs)}!")
        
    # forming x-axis in pixels for plot
    lineoutXPxs = np.arange(len(lineoutY))
    
    # plotting
    if plotsFlag:
        plt.plot(lineoutXPxs, lineoutY)
        plt.axvline(idxLower, color='red', ls='--')
        plt.axvline(idxUpper, color='red', ls='--')
        plt.axvline(center, color='red', ls='--')
        plt.xlabel("Radial (px)")
        plt.ylabel('Intensity (arb.)')
        plt.show()
    return center


def diameterPx(minimaIdxs):
    r"""
    Obtains the diameter of the tube in pixels when given the tube
    edges. Tube edges are obtained using lineoutMinima().
    
    minimaIdxs : tuple
        A tuple of ints corresponding to the pixel positions of the
        two minima in intensity found in the radiograph by lineoutMinima().
        These minima correspond to the inner edge of the target tube.
    """
    # checking that two minima were passed
    if np.shape(minimaIdxs) == (2,):
        # finding the lower and higher pixel values
        idxLower = np.min(minimaIdxs) 
        idxUpper = np.max(minimaIdxs)
    else:
        raise Exception (f"Expected 2 minima, but got {len(minimaIdxs)}!")
    # getting the tube diameter in pixels
    innerDiameterPx = idxUpper - idxLower
    return innerDiameterPx


def magnificationUmPx(innerDiameterPx,
                      stdDevPx=0,
                      innerDiameterUm=500,
                      innerDiameterErrUm=0,
                      printFlag=False):
    r"""
    Get magnification with propagated uncertainty in um/px when given
    target tube diameter in um, in px, and standard deviation on the
    diameter in px.
    
    innerDiameterPx : float
        Measured inner diameter of the target tube in pixels.
        
    stdDevPx : float
        Standard deviation on the inner diameter in pixels. Default is
        zero, for no uncertainty.
    
    innerDiameterUm : float
        Known inner diameter of the target tube in micrometers.
        Default is set to 500 um, which was the nominal diameter for a Marble
        VC target.
        
    innerDiameterErrUm : float
        The error/uncertainty (one sigma) on the inner dimaeter of the
        target tube in micrometers. Default is 0 um.
        
    printFlag : bool
        Flag for printing the calculated magnification. Default is False.
    """
     # getting the magnification (um/px)
    umPerPx = innerDiameterUm / innerDiameterPx
    # error terms for uncertainty propagation based on partial derivatives
    umErrTerm = (innerDiameterPx ** -1) ** 2 * (innerDiameterErrUm) ** 2
    pxErrTerm = (umPerPx / innerDiameterPx) ** 2 * (stdDevPx) ** 2
    # uncertainty on magnification (um/px)
#    stdDevUmPx = np.abs(stdDevPx * umPerPx / innerDiameterPx)
    stdDevUmPx = np.sqrt(umErrTerm + pxErrTerm)
    # printing the magnification, if requested by user
    if printFlag:
        print(f"Magnification is {umPerPx:.3f} +/- {stdDevUmPx:.3f} um/px")
    return umPerPx, stdDevUmPx


#%% BIG BAD MAGNIFICATION ANALYSIS FUNCTION

def magnificationAnalysis(shot,
                          camera,
                          dataDir,
                          px2Um,
                          pxRef,
                          umRef,
                          xMin,
                          xMax,
                          yMin,
                          yMax,
                          cropDownstream=(1000, 1400),
                          innerDiameterUm=500,
                          innerDiameterErrUm=0,
                          saveDir=''):
    r"""
    Get magnification from target radiograph by measuring
    inner tube wall of known diameter.
    
    shot : int
        Shot number to read raw radiographic iamge data from.
        
    camera : str
        Framing camera for which radiographic image to read.
        
    dataDir : str
        Path where data for shot day is stored.
        
    px2Um : float
        Dummy value for magnification conversion to pass to
        initial image loading functions. This just sets the
        axes for those functions.
        
    pxRef : tuple
        Reference position in pixels. This can be a dummy value as
        it is only used by functions loading the image data.
        
    umRef : tuple
        Reference position in micrometers. This can be a dummy value
        as it is only used by functions loading the image data.
        
    xMin : int
        Lower bound pixel for cropping in x-direction.
        
    xMax : int
        Upper bound pixel for cropping in x-direction.
        
    yMin : int
        Lower bound pixel for cropping in y-direction.
        
    yMax : int
        Upper bound pixel for cropping in y-direction.
    
    cropDownstream : tuple
        Tuple of values defining the axial pixel bounds of the downstream
        end of the tube. This is where the tube diameter measurement
        in pixels will be taken. Default is (1000, 1400).
        
    innerDiameterUm : float
        The measured inner diameter of the target tube in
        micrometers. This is used the obtain the magnification in um/px.
        The default is 500 um, which is the nominal value of the Marble VC
        tube.
        
    innerDiameterErrUm : float
        The error/uncertainty (one sigma) on the inner dimaeter of the
        target tube in micrometers. This is used to get the uncertainty
        on the magnification. Default is 0 um.
        
    saveDir : str
        Directory for saving images and data. Default is empty string,
        which does not save the data.
        
    """
    # generating base name for saving iamges and data
    saveStr = saveDir + f'{shot}-{camera}-'
    
    # reading raw image
    imgRaw, _ = shotRadiograph(shotNum=shot,
                               camera=camera,
                               dataDir=dataDir,
                               plots=False)
    
    # crop the image and obtain transforms on spatially calibrated axes
    img_crop, extent_crop, pxRefCrop = cropImgCalibrated(img=imgRaw,
                                                         px2Um=px2Um,
                                                         pxRef=pxRef,
                                                         umRef=umRef,
                                                         xMin=xMin,
                                                         xMax=xMax,
                                                         yMin=yMin,
                                                         yMax=yMax,
                                                         method='horizontal',
                                                         plots=False)
    
    # rotate image to be consistent with Ranjan SBI diagram with shock 
    # propagating downward.
    img_rot, extent_rot, pxRefRot = rotateImgCalibrated(img=img_crop,
                                                        px2Um=px2Um,
                                                        pxRef=pxRefCrop,
                                                        umRef=umRef,
                                                        method='vertical',
                                                        plots=False)
    
    
    # median filtering, morphological filtering, and pseudo-flatfielding
    denoised, cleaned, flattened = enhanceRadiograph(img=img_rot,
                                                     medianDisk=5,
                                                     morphDisk=5,
                                                     flattenMedian=100,
                                                     flattenGauss=100,
                                                     plots=False)
    
    # Selecting downstream region to be used for tube diameter measurement.
    # No need to crop.
    lineoutIdxs = np.arange(start=cropDownstream[0],
                            stop=cropDownstream[1],
                            step=1)
    
    # plotting flatfielded image and downstream region to be used for
    # measuring tube size in pixels
    plt.figure(figsize=(10,12))
    plt.imshow(flattened)
    plt.axhline(cropDownstream[0], color='red', ls='--')
    plt.axhline(cropDownstream[1], color='red', ls='--')
    if saveDir:
        plt.savefig(saveStr + 'flat.png',
                    dpi=300,
                    bbox_inches='tight',
                    pad_inches=0)
    plt.show()
    
    # number of lineouts
    lineoutsNum = cropDownstream[1] - cropDownstream[0]
    
    # initializing arrays for storing tube edges in pixels
    lowerIdxs = np.zeros(lineoutsNum, dtype=int)
    upperIdxs = np.zeros(lineoutsNum, dtype=int)
    centerIdxs = np.zeros(lineoutsNum, dtype=float)
    minimaArr = np.zeros(lineoutsNum, dtype=object)
    
    # produce x-axis array, as this radial axis should be shared by all
    # lineouts in the loop.
    lineoutXFlat = np.arange(np.shape(flattened)[1])
    
    # produce an averaged lineout. This isn't used in the analysis,
    # but is returned to the user for enabling comparisons.
    allLineouts = np.array([flattened[idx, :] for idx in lineoutIdxs])
    lineoutAvg = np.mean(allLineouts, axis=0)
    
    # looping over lineouts
    for idx, lineoutIdx in enumerate(lineoutIdxs):
        # taking lineout through flattened image
        lineoutYFlat = flattened[lineoutIdx, :]
        # finding tube edges. If lineoutX is in units of pixels, then
        # the returned values minimaIdxs and minimaX should be identical.
        minimaIdxs, minimaX, minimaY = lineoutMinima(lineoutX=lineoutXFlat,
                                                     lineoutY=lineoutYFlat,
                                                     window_length=51,
                                                     polyOrder=3,
                                                     peakWidth=40,
                                                     plotsFlag=False)
        # shifting coordinates from being with respect to origin of the
        # cropped image to being with respect to the uncropped image.
        # We shift by yMin, because the image was cropped and then rotated
        # prior to locating tube edges.
        minimaIdxs = minimaIdxs + yMin
        
        # checking that two minima were passed
        if np.shape(minimaIdxs) == (2,):
            # finding the lower and higher pixel values
            idxLower = np.min(minimaIdxs) 
            idxUpper = np.max(minimaIdxs)
        else:
            raise Exception (f"Expected 2 minima, but got {len(minimaIdxs)}!")
        # Finding center pixel position
        centerPx = tubeCenterPx(minimaIdxs=minimaIdxs,
                                lineoutY=lineoutYFlat,
                                plotsFlag=False)
        # recording tube edges and center to arrays for storage
        lowerIdxs[idx] = idxLower
        upperIdxs[idx] = idxUpper
        centerIdxs[idx] = centerPx
        minimaArr[idx] = minimaIdxs
    
    # checking mean, median, and std dev of the wall
    lowerIdxMean = np.mean(lowerIdxs)
    lowerIdxMedian = np.median(lowerIdxs)
    lowerIdxStdDev = np.std(lowerIdxs)
    
    print(f"Lower Wall Mean: {lowerIdxMean:.3f} pixels")
    print(f"Lower Wall Median: {lowerIdxMedian:.3f} pixels")
    print(f"Lower Wall Std. Dev.: {lowerIdxStdDev:.3f} pixels")
    
    # plotting a histogram of the lower wall pixel position
    nLwr, binsLwr, patchesLwr = plt.hist(lowerIdxs,
                                         bins=20)
    plt.axvline(lowerIdxMean, color='red', ls='--')
    plt.axvline(lowerIdxMean - lowerIdxStdDev, color='red', ls='--')
    plt.axvline(lowerIdxMean + lowerIdxStdDev, color='red', ls='--')
    plt.ylabel("# of pts")
    plt.xlabel("Lower wall position (px)")
    if saveDir:
        plt.savefig(saveStr + 'hist-lower-positions.png',
                    dpi=300,
                    bbox_inches='tight',
                    pad_inches=0)
    plt.show()
    
    upperIdxMean = np.mean(upperIdxs)
    upperIdxMedian = np.median(upperIdxs)
    upperIdxStdDev = np.std(upperIdxs)
    
    print(f"Upper Wall Mean: {upperIdxMean:.3f} pixels")
    print(f"Upper Wall Median: {upperIdxMedian:.3f} pixels")
    print(f"Upper Wall Std. Dev.: {upperIdxStdDev:.3f} pixels")
    
    # plotting a histogram of the upper wall pixel position
    nUpr, binsUpr, patchesUpr = plt.hist(upperIdxs,
                                         bins=20)
    plt.axvline(upperIdxMean, color='red', ls='--')
    plt.axvline(upperIdxMean - upperIdxStdDev, color='red', ls='--')
    plt.axvline(upperIdxMean + upperIdxStdDev, color='red', ls='--')
    plt.ylabel("# of pts")
    plt.xlabel("Upper wall position (px)")
    if saveDir:
        plt.savefig(saveStr + 'hist-upper-positions.png',
                    dpi=300,
                    bbox_inches='tight',
                    pad_inches=0)
    plt.show()
    
    # getting tube diameter in pixels for each lineout
    innerDiameterPx = [diameterPx(minimaIdxs) for minimaIdxs in minimaArr]
        
    # Getting mean, median, and standard deviation of tube diameter in pixels
    # for the selected crop region.
    innerDiameterPxMean = np.mean(innerDiameterPx)
    innerDiameterPxMedian = np.median(innerDiameterPx)
    innerDiameterPxStdDev = np.std(innerDiameterPx)
    
    # propagated mean diameter and standard deviation from mean lower
    # and upper Idxs (of tube walls) for comparison
    innerDiameterPxMeanWalls = upperIdxMean - lowerIdxMean
    innerDiameterPxMeansWallErr = np.sqrt(upperIdxStdDev ** 2 + lowerIdxStdDev ** 2)
    
    print(f"Diameter Mean: {innerDiameterPxMean:.3f} pixels")
    print(f"Diameter Median: {innerDiameterPxMedian:.3f} pixels")
    print(f"Diameter Std. Dev.: {innerDiameterPxStdDev:.3f} pixels")
    print(f"Min Diameter in list: {np.min(innerDiameterPx):.3f} pixels")
    print(f"Max Diameter in list: {np.max(innerDiameterPx):.3f} pixels")
    print(f"Diameter propagated from mean edges: {innerDiameterPxMeanWalls:.3f} pixels")
    print(f"Propagated error on diameter from mean edges: {innerDiameterPxMeansWallErr:.3f} pixels")
    
    # plotting a histogram of the measured diameters in pixels
    n, bins, patches = plt.hist(innerDiameterPx,
                                bins=40)
    plt.axvline(innerDiameterPxMean, color='red', ls='--')
    plt.axvline(innerDiameterPxMean - innerDiameterPxStdDev, color='red', ls='--')
    plt.axvline(innerDiameterPxMean + innerDiameterPxStdDev, color='red', ls='--')
    plt.ylabel("# of pts")
    plt.xlabel("Diameter (px)")
    if saveDir:
        plt.savefig(saveStr + 'hist-diameters.png',
                    dpi=300,
                    bbox_inches='tight',
                    pad_inches=0)
    plt.show()
        
    # getting magnification (um/px) and its standard deviation
    umPerPx, stdDevUmPx = magnificationUmPx(innerDiameterPx=innerDiameterPxMean,
                                            stdDevPx=innerDiameterPxStdDev,
                                            innerDiameterUm=innerDiameterUm,
                                            innerDiameterErrUm=innerDiameterErrUm,
                                            printFlag=False)
    print(f"Magnification: {umPerPx:.3f} +/- {stdDevUmPx:.3f} um/px")
    
    # getting mean position of tube center with error
    centerPxMean = np.mean(centerIdxs)
    centerPxMedian = np.median(centerIdxs)
    centerPxStdDev = np.std(centerIdxs)
    
    # propagating center of tube from mean edges of tube for comparison
    centerPxMeanWalls = (upperIdxMean + lowerIdxMean) / 2
    centerPxMeanWallsErr = 0.5 * np.sqrt(upperIdxStdDev ** 2 + lowerIdxStdDev ** 2)
    
    print(f"Center Mean: {centerPxMean:.3f} pixels")
    print(f"Center Median: {centerPxMedian:.3f} pixels")
    print(f"Center Std. Dev.: {centerPxStdDev:.3f} pixels")
    print(f"Center propagated from mean edges: {centerPxMeanWalls:.3f} pixels")
    print(f"Propagated error on center from mean edges: {centerPxMeanWallsErr:.3f} pixels")
        
    # plotting tube edges and center to see if there is any systematic
    # pattern or if variations in position are randomly distributed.
    plt.scatter(lowerIdxs, lineoutIdxs)
    plt.scatter(upperIdxs, lineoutIdxs)
    plt.scatter(centerIdxs, lineoutIdxs)
    # center line with error bounds
    plt.axvline(centerPxMean, color='red', ls='--')
    plt.axvline(centerPxMean + centerPxStdDev, color='red', ls='-')
    plt.axvline(centerPxMean - centerPxStdDev, color='red', ls='-')
    # right wall with error bounds
    plt.axvline(centerPxMean + innerDiameterPxMean/2, color='red', ls='--')
    plt.axvline(upperIdxMean - upperIdxStdDev, color='red', ls='-')
    plt.axvline(upperIdxMean + upperIdxStdDev, color='red', ls='-')
    # left wall with error bounds
    plt.axvline(centerPxMean - innerDiameterPxMean/2, color='red', ls='--')
    plt.axvline(lowerIdxMean - lowerIdxStdDev, color='red', ls='-')
    plt.axvline(lowerIdxMean + lowerIdxStdDev, color='red', ls='-')
    plt.xlabel("Radial (px)")
    plt.ylabel("Axial (px)")
    
    
    # producing a radial coordinate array for the averaged lineout
    lineoutRadii = umPerPx * (lineoutXFlat + yMin - centerPxMean)
    
    if saveDir:
        plt.savefig(saveStr + 'tube-positions.png',
                    dpi=300,
                    bbox_inches='tight',
                    pad_inches=0)
    plt.show()
        
    # saving analysis to dict
    results = {'umPerPx' : umPerPx,
               'stdDevUmPx' : stdDevUmPx,
               'centerPxMean' : centerPxMean,
               'centerPxStdDev' : centerPxStdDev,
               'centerPxMeanWalls' : centerPxMeanWalls,
               'centerPxMeanWallsErr' : centerPxMeanWallsErr,
               'lowerIdxs': lowerIdxs,
               'upperIdxs' : upperIdxs,
               'centerIdxs' : centerIdxs,
               'lineoutIdxs' : lineoutIdxs,
               'innerDiameterPxMean' : innerDiameterPxMean,
               'innerDiameterPxStdDev' : innerDiameterPxStdDev,
               'innerDiameterPxMeanWalls' : innerDiameterPxMeanWalls,
               'innerDiameterPxMeansWallErr' : innerDiameterPxMeansWallErr,
               'lineoutAvg' : lineoutAvg,
               'lineoutRadii' : lineoutRadii}
    
    # returning important values
    return results


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Thu Aug  1 13:13:52 2019

Functions for cleaning and regularizing images.

@author: Pawel M. Kozlowski
"""

# python modules
import numpy as np
from skimage.morphology import disk, opening, closing
from skimage.filters import median, gaussian
import matplotlib.pyplot as plt

# custom modules
import xripl.pltDefaults


def cleanArtifacts(image,
                   diskSize=5,
                   plots=False,
                   vmin=0,
                   vmax=1,
                   flip=False):
    """
    Applies morphological opening to clean bright streaks and spots from
    radiographs, and then applies a morphological closing to remove dark
    artifacts.
    
    image: numpy.ndarray
        Image containing artifacts to be cleaned.
        
    diskSize: int
        Size of disk to be used as convolving element for removing artifcats.
        
    plots: bool
        Flag for displaying plots of processed images.
        
    vmin: float
        Value between 0 and 1 for setting minimum threshold in plotting image.
        See matplotlib.pyplot.imshow().
        
    vmax: float
        Value between 0 and 1 for setting maximum threshold in plotting image.
        See matplotlib.pyplot.imshow().
        
    flip: bool
        Flips order of opening and closing operations. When False, opening
        is applied first, then closing. When True, closing is applied first,
        then opening. False is used by default for radiographs, and True
        is used for inverted radiographs
    """
    if flip:
        imgClosed = closing(image, disk(diskSize))
        cleaned = opening(imgClosed, disk(diskSize))
    else:
        imgOpen = opening(image, disk(diskSize))
        cleaned = closing(imgOpen, disk(diskSize))
    
    if plots:
        fig, axes = plt.subplots(nrows=1,
                                 ncols=3,
                                 figsize=(12, 12),
                                 sharex=True,
                                 sharey=True)
        ax = axes.ravel()
        
        ax[0].imshow(image, vmin=vmin, vmax=vmax)
        ax[0].set_title("Original")
        
        if flip:
            ax[1].imshow(imgClosed)
            ax[1].set_title("Closed")
            
            ax[2].imshow(cleaned)
            ax[2].set_title("Opened")
        else:
            ax[1].imshow(imgOpen)
            ax[1].set_title("Opened")
            
            ax[2].imshow(cleaned)
            ax[2].set_title("Closed")
        
        fig.tight_layout()
        plt.show()
    return cleaned


def flatten(image, medianDisk=100, gaussSize=100, plots=False, vmin=0, vmax=1):
    """
    Produces an approximate image of the background light intensity
    variations in the image and divides the image by this background to
    effectively flatten the contrast.
    
    image: numpy.ndarray
        Image whose background brightness variation is to be removed via
        flattening.
        
    medianDisk: int
        Size of disk used in median filtering to smooth out all structures
        in image other than background light variations.
        
    gaussSize: int
        Size of Gaussian smoothing used on background image produced by
        median filtering. This smooths out ridges that appear in the
        median filtered image.
        
    plots: bool
        Flag for displaying plots of processed images.
    """
    
    # removing background from image
    denoised = median(image, disk(medianDisk))
    # smoothing background
    gaussed = gaussian(denoised, sigma=gaussSize)
    # renaming
    pseudoFlatfield = gaussed
    # applying pseudo flatfield
    flat = image / pseudoFlatfield
    # filtering nan and inf values
    flat[np.isnan(flat)] = 0
    flat[np.isinf(flat)] = 0
    # reshaping from flat array to 2D array after filtering out nan values
    imgShape = np.shape(image)
    flat = flat.reshape(imgShape)
    # normalizing
    flat_norm = flat / np.max(flat)
    
    if plots:
        fig, axes = plt.subplots(nrows=1,
                                 ncols=3,
                                 figsize=(12, 12),
                                 sharex=True,
                                 sharey=True)
        ax = axes.ravel()
        
        ax[0].imshow(image, vmin=vmin, vmax=vmax)
        ax[0].set_title("Original")
        
        ax[1].imshow(pseudoFlatfield, vmin=vmin, vmax=vmax)
        ax[1].set_title("Background")
        
        ax[2].imshow(flat_norm, vmin=vmin, vmax=vmax)
        ax[2].set_title("Flattened")
        
        fig.tight_layout()
        plt.show()
    return flat_norm


def enhanceRadiograph(img,
                      medianDisk=5,
                      morphDisk=5,
                      flattenMedian=100,
                      flattenGauss=100,
                      plots=False):
    r"""
    Median filter, morphological filter, and pseudo-flatfield the
    radiographs to prepare it for feature identification.
    
    img : numpy.ndarray
        Foreground radiographic image as a 2D numpy array, which is
        to be cleaned/enhanced.
        
    medianDisk : int
        Radial size of disk in pixels to be used for median filtering
        the image. Default is radius of 5 pixels.
        
    morphDisk : int
        Radial size of disk in pixels to be used for morphologically
        filtering the image to clean up artifacts.
        
    flattenMedian : int
        Radial size of disk in pixels to be used for median filtering
        image as part of over-smoothing process to obtain the
        pseudo-flatfield. Default is radius of 100 pixels.
        
    flattenGauss : int
        Standard deviation in pixels to be used for Gaussian blurring
        the image as part of over-smoothing process to obtain the
        pseudo-flatfield. Default is sigma of 100 pixels.
        
    plots : bool
        Flag for plotting cleaned and flattened images. Default is False.
    """
    # normalizing image intensity to max value
    img_norm = img / np.max(img)
    # median filtering to denoise the image
    denoised = median(img_norm, disk(medianDisk))
    denoised_norm = denoised / np.max(denoised)
    # removing artifacts (e.g. xrfc streaks, hot pixels, etc.)
    cleaned = cleanArtifacts(image=denoised_norm,
                             diskSize=morphDisk,
                             plots=plots)
    # Flattening image by dividing out approxiamte background
    flattened = flatten(image=cleaned,
                        medianDisk=flattenMedian,
                        gaussSize=flattenGauss,
                        plots=plots)
    
    # returns median filtered image, morphologically filtered image
    # and pseudo-flatfielded image as a tuple.
    return denoised, cleaned, flattened


# -*- coding: utf-8 -*-
"""
Created on Mon Oct 22 16:42:16 2018

Reads radiographic images and create simple plots

@author: Pawel M. Kozlowski
"""

# python modules
import glob
import os
import numpy as np
import h5py as h5
import matplotlib.pyplot as plt

# custom modules
import xripl.pltDefaults


def openRadiograph(fileName):
    """
    Given the full filename (with path) to an hdf5 file containing
    radiographic data from an XRFC (x-ray framing camera) diagnostic
    on OMEGA, open the file and return the foreground and background
    image data as numpy arrays.
    
    fileName is a str
    """
    with h5.File(fileName, "r") as f:
        streakArr = f["Streak_array"][...]
    
    foreground = streakArr[0, :, :]
    background = streakArr[1, :, :]
    return foreground, background


def omegaDataSearch(dataDir, camera='xrfc4'):
    """
    Search given directory and subdirectories for HDF5 files of
    radiographs from OMEGA LLE XRFC data.
    """
    if not camera in ['xrfc1', 'xrfc2', 'xrfc3', 'xrfc4', 'xrfc5']:
        raise Exception(f"Cannot find method for camera {camera}")
        
    # fetch all xrfccd hdf5 files in directory and subdirectories
    pattern = f'**/xrfccd_{camera}*.h5'
    flist = glob.glob(os.path.join(dataDir, pattern), recursive=True)
    # trim the filenames to just base filename
    flistBase = [os.path.basename(fname) for fname in flist]
    # split filename and extension
    flistTrim = [os.path.splitext(fname)[0] for fname in flistBase]
    # extracting shot numbers from file names
    shotNos = [int(fname.split("_")[-1]) for fname in flistTrim]
    # get diagnostic names from prepended portion of filenames
    diagnosticNamesList = ["_".join(fname.split("_")[:-1]) for fname in flistTrim]
    diagnosticNames = set(diagnosticNamesList)
    return flist, shotNos, diagnosticNames, flistTrim


def shotRadiograph(shotNum, camera, dataDir, plots=False):
    r"""
    Opens a radiograph given a shot number, framing camera and
    directory containing shot data.
    
    shotNum : int
        Omega shot number of radiograph to be opened.
        
    camera : str
        Name of camera used on shotNum from which to open radiograph.
        For example, 'xrfc3' for x-ray framing camera 3.
        
    dataDir : str
        Full path to directory containing shot data with radiographs.
        
    plots : bool
        Flag for plotting the background and the foreground images.
        Default is False.
    """
    # get files in directory and pick a shot
    flist, shotNos, diagnosticNames, flistTrim = omegaDataSearch(dataDir=dataDir,
                                                                  camera=camera)
    
    # picking a shot
    if not shotNum in shotNos:
        raise Exception(f"Shot number {shotNum} not found in directory!")
    # looking for shot number in files in directory
    filteredFiles = np.array([str(shotNum) in fileName for fileName in flist])
    # picking just the first match (don't care which diagnostic it is, so long
    # as shot number is correct, since we already filtered based on 
    # camera diagnostic)
    flistArr = np.array(flist)
    xrfcFull = flistArr[filteredFiles][0]
#    baseName = os.path.basename(xrfcFull)
    
    # open the selected radiograph if it is in the list of supported
    # camera types for this data set
    if camera in ['xrfc1', 'xrfc2', 'xrfc3', 'xrfc4', 'xrfc5']:
        foreground, background = openRadiograph(fileName=xrfcFull)
    else:
        raise NotImplementedError(f"Unsupported camera type {camera}.")
    
    if camera == 'xrfc5':
        # mirroring the image so that it is oriented in same direction
        # as XRFC3
        foreground = np.fliplr(foreground)
        background = np.fliplr(background)
    
    if plots:
        plt.imshow(background)
        plt.title(f"Background")
        plt.show()
        
        plt.figure(figsize=(10,12))
        plt.imshow(foreground)
        plt.title(f"Foreground")
        plt.show()
    return foreground, background


# -*- coding: utf-8 -*-
"""
Created on Mon Oct 22 16:46:08 2018

Functions for adjusting plotting and adjusting the contrast
of radiographic images.

@author: Pawel M. Kozlowski
"""

# python modules
#import numpy as np
import matplotlib.pyplot as plt
from skimage import exposure
#from skimage.morphology import disk, square
#from skimage.filters import rank
    
# custom modules
import xripl.pltDefaults



#%% global and CLAHE contrast equalization

def equalize(image,
             plot=False,
             savePlot=False,
             fileName="",
             saveDir="",
             cmap='viridis',
             kernel_size=(256, 256),
             clip_limit=0.03,
             nbins=256):
    """
    Convenience function for applying contrast equalization to radiographic
    data. Applies both global contrast equalization and CLAHE equalization.
    """
    # globally equalized contrast image
    globalImg = exposure.equalize_hist(image)
    
    # CLAHE equalized contrast image
#    claheImg = exposure.equalize_adapthist(image,
#                                          kernel_size=(256,256),
#                                          clip_limit=0.03,
#                                          nbins=256)
    
    claheImg = exposure.equalize_adapthist(image,
                                           kernel_size=kernel_size,
                                           clip_limit=clip_limit,
                                           nbins=nbins)

    # locally equalized contrast image
    #selem = disk(100)
    ##selem = square(1000)
    #locallyEqualized = rank.equalize(foreground, selem=selem)
    
    if plot:
        plt.figure(figsize=(10,10))
        plt.imshow(image, cmap=cmap)
        plt.title(f"{fileName} Unprocessed image")
        plt.show()
        # plot the contrast equalized images along with histograms
        # plotting globally equalized images
        plt.figure(figsize=(10,10))
        plt.imshow(globalImg, cmap=cmap)
        plt.title(f"{fileName} Globally equalized contrast")
        plt.show()
        
        # ravel to flatten image array and plot histogram of intensity
#        plt.hist(globalImg.ravel(), bins=256)
#        plt.title(f"{fileName} Globally equalized histogram")
#        plt.show()
        
        # plotting CLAHE equalized images
        plt.figure(figsize=(10,10))
        plt.imshow(claheImg, cmap=cmap)
        plt.title(f"{fileName} CLAHE equalized")
        plt.show()
        
        # ravel to flatten image array and plot histogram of intensity
#        plt.hist(claheImg.ravel(), bins=256)
#        plt.title(f"{fileName} CLAHE equalized histogram")
#        plt.show()
        
        # plotting locally equalized images
        #plt.imshow(locallyEqualized)
        #plt.title("Locally Equalized Intensity")
        #plt.show()
        #
        ## ravel to flatten image array and plot histogram of intensity
        #plt.hist(locallyEqualized.ravel(), bins=256)
        #plt.title("Locally Equalized intensity histogram")
        #plt.show()
        
    if savePlot:
        plt.figure(figsize=(10,10))
        plt.imshow(image, cmap=cmap)
        plt.title(f"{fileName} Unprocessed image")
        fname = saveDir + fileName + ".jpg"
        plt.savefig(fname=fname)
        # plot the contrast equalized images along with histograms
        # plotting globally equalized images
        plt.figure(figsize=(10,10))
        plt.imshow(globalImg, cmap=cmap)
        plt.title(f"{fileName} Globally equalized contrast")
        fname = saveDir + fileName + "_global.jpg"
        plt.savefig(fname=fname)
        
        # plotting CLAHE equalized images
        plt.figure(figsize=(10,10))
        plt.imshow(claheImg, cmap=cmap)
        plt.title(f"{fileName} CLAHE equalized")
        fname = saveDir + fileName + "_CLAHE.jpg"
        plt.savefig(fname=fname)
        
    return globalImg, claheImg



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Mar 31 09:50:47 2020

Custom filters for cleaning/blurring images.

@author: Pawel M. Kozlowski
"""

import numpy as np
from scipy import ndimage

def boxFilter(img, boxSize=2):
    """
    Applies a symmetric box filter to blur the image. A box filter
    approaches a Gaussian filter for large number of filterings, but
    at smaller kernel sizes it causes less blur. Convolution is
    handled using scipy.ndimage.convolve, and edges condition is 'reflect'.
    
    img : numpy.ndarray
        2D numpy array of the image to be filtered.
        
    boxSize : int
        Length in pixels of the side of the box. The box is symmetric.
        Default size 2 pixels by 2 pixels.
        
    filteredImg : numpy.ndarray
        Box filtered image returned as a 2D numpy array.
    """
    boxKernel = np.ones((boxSize, boxSize)) / boxSize ** 2
    filteredImg = ndimage.convolve(img, boxKernel, mode='reflect')
    return filteredImg


# -*- coding: utf-8 -*-
"""
Created on Mon Oct 22 16:47:46 2018

Tools for identifying contours of equal intensity/contrast.

@author: Pawel M. Kozlowski
"""

# python modules
import numpy as np


def nContours(contours, n):
    """
    Get the n longest contours from a list of contours.
    
    contours: list
        A list of contours from skimage.measure.find_contours().
        
    n: int
        Number of contours to select.
    """
    # getting lengths of contours arrays
    contLens = np.array([len(cont) for cont in contours])
    # sorting indices of contour list based on length of contour
    contIdxSorted = np.argsort(contLens)[::-1]
    # Reducing sorted index list to n longest contours
    longestContoursIdxs = contIdxSorted[:n]
    # filtering contour list using selected indices of n longest contours
    contours_np = np.array(contours)
    longestContours = contours_np[longestContoursIdxs]
    return longestContours


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Fri Oct 27 02:37:12 2017

Default plotting parameters

@author: Pawel M. Kozlowski
"""

# importing Python modules
import numpy as np
import matplotlib 
import matplotlib.pyplot as plt
from itertools import cycle

# setting default plot properties
plotFont = {'family' : 'serif',
            'weight' : 'normal',
            'size'   : 16}
plt.rc('font', **plotFont)
plt.rc('lines', linewidth=2)
plt.rc('lines', markersize=8)
#plt.rc('text', usetex=True)
#plt.rc('text', usetex=False)
# setting default figure size
matplotlib.rcParams['figure.figsize'] = 10, 6

validMathTextFonts = ['dejavusans',
                      'dejavuserif',
                      'cm',
                      'stix',
                      'stixsans',
                      'custom']

# setting latex rendered fonts to be same as regular fonts
try:
    matplotlib.rcParams['mathtext.fontset'] = 'dejavuserif'
    matplotlib.rcParams['mathtext.rm'] = 'DejaVu Serif'
    matplotlib.rcParams['mathtext.it'] = 'DejaVu Serif:italic'
    matplotlib.rcParams['mathtext.bf'] = 'DejaVu Serif:bold'
except:
    print("Couldn't load dejavuserif fonts for plot defaults."
          "Falling back to stix fonts.")
    try:
        matplotlib.rcParams['mathtext.fontset'] = 'stix'
        matplotlib.rcParams['font.family'] = 'STIXGeneral'
    except:
        print("Couldn't load stix fonts for plot defaults.")

# for testing whether matplotlib and python fonts match
#plt.title(r'cm$\rm cm^{-3}$')


#%% convenience functions for plotting

# shaded error bars for line plot
def plot_line_shaded(xData, yData, yErrs, label="", **kwargs):
    """
    Generate a line plot with shaded region representing y-error bars.
    Can be run multiple times before plt.show(), to plot multiple data
    sets on the same axes.
    x axis data points
    y axis data points
    y axis errors
    """
    # check that arrays are 1D
    
    # check that arrays are equal length
    if not len(xData) == len(yData) == len(yErrs):
        raise ValueError("Arrays must of equal length!")
    
    # line plot
    ax = kwargs.pop('ax', plt.gca())
    base_line, = ax.plot(xData, yData, label=label, **kwargs)
    # shaded error region
    plt.fill_between(xData,
                     yData - yErrs,
                     yData + yErrs,
                     alpha=0.5,
                     edgecolor=base_line.get_color(),
                     facecolor=base_line.get_color())
    
    return

# error bars for scatter plot
def plot_scatter_bars(xData, yData, yErrs, label="", **kwargs):
    """
    Generate a scatter plot with y-error bars.
    Can be run multiple times before plt.show(), to plot multiple data
    sets on the same axes.
    x axis data points
    y axis data points
    y axis errors
    """
    # check that arrays are 1D
    
    # check that arrays are equal length
    if not len(xData) == len(yData) == len(yErrs):
        raise ValueError("Arrays must of equal length!")
        
    # line plot
    ax = kwargs.pop('ax', plt.gca())
    ax.errorbar(xData,
                yData,
                yerr=yErrs,
                label=label,
                fmt='o',
                fillstyle='none',
                capsize=4,
                elinewidth=1,
                **kwargs)
    return



#%% style cycling
# printing default colors
default_colors = matplotlib.colors.cnames.keys()
#print(f"Colors: {default_colors}")
# printing default linestyles
default_lines = matplotlib.lines.lineStyles.keys()
#print(f"Lines: {default_lines}")
# printing default marker styles
default_markers = matplotlib.markers.MarkerStyle.markers.keys()
#print(f"Markers: {default_markers}")

# custom list of linestyles (excluding blank line styles)
lines = ["-","--","-.",":"]
linecycler = cycle(lines)
#plt.figure()
#for i in range(10):
#    x = range(i,i+10)
#    plt.plot(range(10),x,next(linecycler))
#plt.show()
    


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue May 12 10:59:55 2020

Idealized estimate of what a 1D lineout radially through
a cylindrical foam + tube would look like. This is to assist
in correctly identifying which features in a radiograph correspond
to which part of the target for spatial magnification calculation.

@author: Pawel M. Kozlowski
"""
# python modules
import numpy as np
import matplotlib.pyplot as plt
import scipy.signal as sig

# custom modules
import xripl.pltDefaults


def chord(radius, height):
    r"""
    Calculates the horizontal chord length through a circle at a given
    height. If the height is larger than the circle's radius, then
    a chord length of zero is returned.
    
    radius : float
        Radius of the circle. Must be in same units as height.
        
    height : float
        Height along the circle at which to calculate the horizontal
        chord length.
    """
    # taking the absolute value of height in case a negative value is
    # given. Since a circle is symmetric, a negative height coordinate
    # is the same as a positive height coordinate.
    height = np.abs(height)
    if radius <= 0:
        raise Exception(f"Circle radius must be positive.")
    if height > radius:
        return 0
    else:
        return 2 * np.sqrt(radius ** 2 - height ** 2)
    
    
def tubeThickness(innerDiameter,
                  outerDiameter,
                  height):
    r"""
    Calculates the chord lengths through the inner region (foam) and outer
    region (tube wall) for a given target. Here the tube is simply
    an anulus with the given inner and outer diameters as dimensions, and
    the region inside the tube is the "foam" region.
    
    innerDiameter : float
        Inner diameter of the tube, which is also the diameter of the foam.
        
    outerDiameter : float
        Outer diameter of the tube. Must be in same units as innerDiameter
        and height.
        
    height : float
        Height at which to calculate the chord lengths through the
        tube and foam.
    """
    radiusInner = innerDiameter / 2
    radiusOuter = outerDiameter / 2
    # getting chord lengths for the inner and outer circles which
    # describe the tube anulus.
    chordInner = chord(radius=radiusInner, height=height)
    chordOuter = chord(radius=radiusOuter, height=height)
    # thickness of the foam at the given height
    lengthFoam = chordInner
    # thickness of just the tube (excluding the foam) at the given height
    lengthTube = chordOuter - chordInner
    return lengthTube, lengthFoam


def transmission(density, opacity, length):
    r"""
    Calculates the transmission through a uniform material of given
    density, opacity, and length.
    
    density : float
        Mass density of the material in grams per cubic centimeter.
        
    opacity : float
        Opacity of the material (also known as mass attenuation coefficinet)
        in squared centimeters per gram.
        
    length : float
        Attenuation length through the material in centimeters.
    """
    return np.exp(-density * opacity * length)


def tubeTransmission(innerDiameter,
                     outerDiameter,
                     densityTube,
                     densityFoam,
                     opacityTube,
                     opacityFoam,
                     height):
    r"""
    Calculates the transmission through a tube with foam at a given
    height (radial coordinate) through the tube.
    
    innerDiameter : float
        Inner diameter of the tube, which is also the diameter of the foam.
        In units of centimeters.
        
    outerDiameter : float
        Outer diameter of the tube. In units of centimeters.
        
    densityTube : float
        Mass density of the tube material in grams per cubic centimeter.
        
    densityFoam : float
        Mass density of the foam material in grams per cubic centimeter.
        
    opacityTube : float
        Opacity of the tube material (aka mass attenuation coefficient)
        in squared centimeters per gram.
        
    opacityFoam : float
        Opacity of the foam material (aka mass attenuation coefficient)
        in squared centimeters per gram.
        
    height : float
        Height at which to calculate the chord lengths through the
        tube and foam. In units of centimeters.
    """
    # get thickness of tube wall and foam respectively for the given
    # height. These are chord lengths through an anulus.
    lengthTube, lengthFoam = tubeThickness(innerDiameter=innerDiameter,
                                           outerDiameter=outerDiameter,
                                           height=height)
    # transmission through tube
    transTube = transmission(density=densityTube,
                             opacity=opacityTube,
                             length=lengthTube)
    # transmission through foam
    transFoam = transmission(density=densityFoam,
                             opacity=opacityFoam,
                             length=lengthFoam)
    # total transmission is just the multiplication of the two
    # transmissions.
    transTotal = transTube * transFoam
    return transTotal


def plotTubeTransmission(innerDiameter,
                         outerDiameter,
                         densityTube,
                         densityFoam,
                         opacityTube,
                         opacityFoam,
                         heightsArr,
                         plots=True):
    r"""
    Generates a plot transmission versus radial coordinate through
    a tube.
    
    innerDiameter : float
        Inner diameter of the tube, which is also the diameter of the foam.
        In units of centimeters.
        
    outerDiameter : float
        Outer diameter of the tube. In units of centimeters.
        
    densityTube : float
        Mass density of the tube material in grams per cubic centimeter.
        
    densityFoam : float
        Mass density of the foam material in grams per cubic centimeter.
        
    opacityTube : float
        Opacity of the tube material (aka mass attenuation coefficient)
        in squared centimeters per gram.
        
    opacityFoam : float
        Opacity of the foam material (aka mass attenuation coefficient)
        in squared centimeters per gram.
        
    heightsArr : numpy.ndarray
        Array of heights along the tube radial axis at which to calculate the
        chord lengths through the tube and foam for getting transmission.
        These can be negative and positive. In units of centimeters.
        
    plots : bool
        Flag for generating plots. Default is True.
    """
    innerRadius = innerDiameter / 2
    outerRadius = outerDiameter / 2
    # array for storing transmissions. Default value is 1, for no
    # attenuation through the target.
    transmissions = np.ones_like(heightsArr)
    for idx, height in enumerate(heightsArr):
        transmissions[idx] = tubeTransmission(innerDiameter=innerDiameter,
                                              outerDiameter=outerDiameter,
                                              densityTube=densityTube,
                                              densityFoam=densityFoam,
                                              opacityTube=opacityTube,
                                              opacityFoam=opacityFoam,
                                              height=height)
    
    # normalizing heights to inner radius of tube
    heightsArrNormInner = heightsArr / innerRadius
    # normalizing heights to outer radius of tube
    heightsArrNormOuter = heightsArr / outerRadius
    
    if plots:
        # transmission vs height plot
        plt.plot(heightsArr, transmissions)
        plt.axvline(innerRadius, color='red', ls='--')
        plt.axvline(outerRadius, color='red', ls='--')
        plt.xlabel('Radial height (cm)')
        plt.ylabel('Transmission')
        plt.show()
        
        # transmission vs normalized height plot (inner radius)
        plt.plot(heightsArrNormInner, transmissions)
        plt.axvline(1, color='red', ls='--')
        plt.xlabel('Height norm inner')
        plt.ylabel('Transmission')
        plt.show()
        
        # transmission vs normalized height plot (outer radius)
        plt.plot(heightsArrNormOuter, transmissions)
        plt.axvline(1, color='red', ls='--')
        plt.xlabel('Height norm outer')
        plt.ylabel('Transmission')
        plt.show()
    return heightsArr, transmissions, heightsArrNormInner, heightsArrNormOuter


def gaussianBlur(heightsArr,
                 transmissions,
                 stdDevCm,
                 innerDiameter,
                 outerDiameter,
                 plots=False):
    r"""
    
    heightsArr : numpy.ndarray
        Array of heights along the tube radial axis at which chord lengths
        through the tube and foam were used for getting transmission.
        These can be negative and positive. In units of centimeters.
        See plotTubeTransmission().
        
    transmissions : numpy.ndarray
        Transmission values through the tube corresponding to heightsArr.
        See plotTubeTransmission().
        
    stdDevCm : float
        Standard deviation of Gaussian (in centimeters) to be used
        for blurring the transmissions curve.
        
    innerDiameter : float
        Inner diameter of the tube, which is also the diameter of the foam.
        In units of centimeters. Used to overlay tube location on plots.
        
    outerDiameter : float
        Outer diameter of the tube. In units of centimeters. Used to
        overaly tube location on plots.
        
    plots : bool
        Flag for plotting Gaussian used in convolution and blurred
        transmissions curve.
    """
    # Get length of input array to form Gaussian
    lengthArr = len(heightsArr)
    # convert stdDev from cm to pixels
    stdDevPx = int(stdDevCm * (lengthArr / np.max(heightsArr)))
    # normalization factor to make integral of Gaussian equal to 1
    normFactor = 1 / (stdDevPx * np.sqrt(2 * np.pi))
    # forming Gaussian curve
    gaussian = normFactor * sig.gaussian(lengthArr, std=stdDevPx) 
    
    # Padding array edges to avoid edge effects due to convolution.
    # This pads the array with edge values instead of zeros.
    transmissionsPad = np.pad(transmissions,
                              (lengthArr, lengthArr),
                              'edge')
    
    # convolving gaussian with transmission array
    transmissionsBlurPad = sig.convolve(transmissionsPad, 
                                        gaussian,
                                        mode='same')
    
    # removing padded regions to bring length of array back to
    # original length
    transmissionsBlur = transmissionsBlurPad[lengthArr:-lengthArr]
    
    
    if plots:
        # plotting the Gaussian
        plt.plot(heightsArr - np.max(heightsArr)/2, gaussian)
        plt.xlabel('Radial (cm)')
        plt.title('Gaussian for blurring')
        plt.show()
        
        # plotting blurred transmission
        plt.plot(heightsArr, transmissionsBlur)
        # overlaying tube edge locations
        plt.axvline(innerDiameter/2, color='red', ls='--')
        plt.axvline(outerDiameter/2, color='red', ls='--')
        plt.xlabel('Radial height (cm)')
        plt.ylabel('Transmission Blurred')
        plt.ylim(0.55, 1.05)
        plt.show()
    
    return transmissionsBlur





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Wed Feb 27 15:32:34 2019

Opening and processing microD scanned films.

@author: Pawel M. Kozlowski
"""

# python modules
import h5py as h5


#%% functions

def openFilm(fileName):
    """
    Opens HDF5 file film scan from microD.
    """
    with h5.File(fileName, "r") as f:
        stuff = f['pds_image'][...]
    return stuff





#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Feb 26 16:11:09 2019

XRIPL setup script.

@author: Pawel M. Kozlowski
"""

from setuptools import setup

setup(name='xripl',
      version='0.1',
      description='X-Ray Radiographic Image Processing Library',
      url='https://github.com/lanl/xripl',
      author='Pawel Marek Kozlowski',
      author_email='pkozlowski@lanl.gov',
      license='BSD',
      classifiers=['Intended Audience :: Science/Research',
                   'License :: OSI Approved :: BSD License',
                   'Development Status :: 3 - Alpha',
                   'Operating System :: OS Independent',
                   'Programming Language :: Python :: 3',
                   'Topic :: Scientific/Engineering :: Physics',
                   'Topic :: Scientific/Engineering :: Image Processing',
                   'Natural Language :: English'],
      packages=['xripl'],
      zip_safe=False)


#!/usr/bin/env python3
# -*- coding: utf-8 -*-
"""
Created on Tue Feb 26 12:58:48 2019

Utilities for convolving 2D radiograph instrument
function with synthetically produced radiographs.

@author: Pawel M. Kozlowski
"""

# python modules
import os
import numpy as np
import h5py as h5
import matplotlib.pyplot as plt
import skimage
from skimage import filters

# custom modules
import xripl.pltDefaults



#%% functions
def openSyntheticRadiograph(directory, fileName, plot=False):
    """
    Given the full filename (with path) to an hdf5 file containing
    synthetic radiograph data, open the file and return the image data as
    a numpy array.
    
    fileName is a str
    """
    with h5.File(directory + fileName, "r") as f:
        rasterImg = f['Raster Image #0'][...].transpose()
    
    if plot:
        plt.imshow(rasterImg)
        plt.title(fileName)
        plt.show()
    return rasterImg


def saveHdf(data, directory, fileName):
    """
    Saves radiograph into HDF5 file.
    """
    fullName = directory + fileName
    with h5.File(fullName, "w") as f:
        f.create_dataset("Raster Image #1", data=data)
    return


